#include <LiquidCrystal.h>
#define tic 0x927C0

#define EN_PIN   4 // Пин включения драйвера
#define STEP_PIN  3 // Пин сигнала шага драйвера (CLC, STEP)
#define DIR_PIN  2 //Пин направления вращения (CW, DIR)

#define EXT_C  19 // Пин входа аналогового управления
#define EXTR_ON 18// Пин входа внешнего старт/стопа
#define KEY_C 13  // Пин тумблера внешнего управления

//Тестовая раскладка ПИНов
#define RS_PIN 5 // Пин сигнала RES дисплея
#define RW_PIN 6 // Пин сигнала RW дисплея
#define E_PIN 12 // Пин сигнала En дисплея
#define D4_PIN 10 // Пин полубайта данных дисплея
#define D5_PIN 11 // Пин полубайта данных дисплея
#define D6_PIN 8 // Пин полубайта данных дисплея
#define D7_PIN 9 // Пин полубайта данных дисплея

#define KEY_1 17 // Пин кнопки 1
#define KEY_2 16 // Пин кнопки 2
#define KEY_3 19 // Пин кнопки 3
#define KEY_4 18 // Пин кнопки 4

//Раскладка ПИНов по схеме
//#define RS_PIN 5 // Пин сигнала RES дисплея
//#define RW_PIN 6 // Пин сигнала RW дисплея
//#define E_PIN 12 // Пин сигнала En дисплея
//#define D4_PIN 8 // Пин полубайта данных дисплея
//#define D5_PIN 9 // Пин полубайта данных дисплея
//#define D6_PIN 10 // Пин полубайта данных дисплея
//#define D7_PIN 11 // Пин полубайта данных дисплея

//#define KEY_1 14 // Пин кнопки 1
//#define KEY_2 15 // Пин кнопки 2
//#define KEY_3 16 // Пин кнопки 3
//#define KEY_4 17 // Пин кнопки 4

#define ANALOG_KEYBOARD 0 // 0 - цифровые кнопки | 1 - аналоговая клавиатура LCD Shield
#define KEY_ANALOG 0 // Пин аналоговой клавиатуры шилда, меняем под свой вариант, по умолчанию A0

#define LCD_ROW 2 // Строк дисплея
#define LCD_COL 16 // Символов в строке

//Настройка соответствия аналоговых кнопок
#define bRIGHT  0
#define bUP     1
#define bDOWN   2
#define bLEFT   3
#define bSELECT 4
#define bNONE   5

LiquidCrystal lcd(RS_PIN, RW_PIN, E_PIN, D4_PIN, D5_PIN, D6_PIN, D7_PIN); //Конфигурируем дисплей

volatile uint16_t count;
volatile unsigned long speed;
bool run;
unsigned int adc_key_in;
bool start, ext_on, ext_on_prev, dir, key[4], key_action[4], refresh_speed, refresh_dir, refresh_run, refresh_disp, refresh_cont, info, refresh_info, control_enable, control_tmp;
float max_rpm;

char info_1[LCD_COL] = {'P', 'e', 'r', 'i', 's', 't', 'a', 'l', 't', 'i', 'c', ' ', 'p', 'u', 'm', 'p'};
char info_2[LCD_COL] = {' ', ' ', 'V', 'e', 'r', 's', 'i', 'o', 'n', ' ', '0', '.', '9', '6', 'T', '2'};

byte s1[8] = {0b00100, 0b00100, 0b00100, 0b00100, 0b00100, 0b11111, 0b01110, 0b00100};
byte s2[8] = {0b00100, 0b01110, 0b11111, 0b00100, 0b00100, 0b00100, 0b00100, 0b00100};

char rpm_str[5], adc_str[5];

float rpm, rpm_save, external, external_old;
unsigned long key_time[5];

void setup()
{
  lcd.createChar(0, s1);
  lcd.createChar(1, s2);

  lcd.begin(LCD_COL, LCD_ROW);

  pinMode(STEP_PIN , OUTPUT);
  pinMode(DIR_PIN , OUTPUT);
  pinMode(EN_PIN , OUTPUT);

  digitalWrite(DIR_PIN , HIGH);

  external = 0;
  max_rpm = 300;
  rpm = 100;
  rpm_save = rpm;
  dir = false;
  speed = tic / rpm;
  run = false;
  start = false;
  info = true;
  refresh_speed = true;
  refresh_dir = true;
  refresh_run = true;
  refresh_info = true;
  pinMode(KEY_C, INPUT);
  pinMode(EXTR_ON,INPUT);
  if (ANALOG_KEYBOARD == 0)
  {
    pinMode(KEY_1, INPUT);
    pinMode(KEY_2, INPUT);
    pinMode(KEY_3, INPUT);
    pinMode(KEY_4, INPUT);
  }
  TCCR2A = (1 << WGM21);
  TCCR2B = (1 << CS20);
  OCR2A = 0x80;
  TIMSK2 = (1 << OCIE2A);

  sei ();
}

void read_ext_control()
{
  external = analogRead(EXT_C);
  ext_on = digitalRead(EXTR_ON);  
}

int read_analog_buttons()
{
  adc_key_in = analogRead(KEY_ANALOG);
  if (adc_key_in > 1000) return bNONE;
  if (adc_key_in < 50)   return bRIGHT;
  if (adc_key_in < 250)  return bUP;
  if (adc_key_in < 450)  return bDOWN;
  if (adc_key_in < 650)  return bLEFT;
  if (adc_key_in < 850)  return bSELECT;
  return bNONE;
}

void key_read()
{
  int key_select;
  control_tmp = digitalRead(KEY_C);
  if (control_tmp != control_enable)
  {
    run = false;
    if (control_enable) 
     { 
      rpm = rpm_save;
      ext_on_prev=false;
     }
    else rpm_save = rpm;
    control_enable = control_tmp;
    refresh_cont = true;
    refresh_speed = true;
    refresh_run = true;
  }
  if (ANALOG_KEYBOARD == 0)
  {
    key[0] = digitalRead(KEY_1);
    key[1] = digitalRead(KEY_2);
    key[2] = digitalRead(KEY_3);
    key[3] = digitalRead(KEY_4);
  }
  else
  {
    key_select = read_analog_buttons();

    if (key_select == bSELECT)
    {
      key[0] = true;
      key[1] = false;
      key[2] = false;
      key[3] = false;
    }
    if (key_select == bLEFT)
    {
      key[0] = false;
      key[1] = true;
      key[2] = false;
      key[3] = false;
    }
    if (key_select == bUP)
    {
      key[0] = false;
      key[1] = false;
      key[2] = true;
      key[3] = false;
    }
    if (key_select == bRIGHT)
    {
      key[0] = false;
      key[1] = false;
      key[2] = false;
      key[3] = true;
    }
    if (key_select == bNONE)
    {
      key[0] = false;
      key[1] = false;
      key[2] = false;
      key[3] = false;
    }
  }
}

void key_check()
{
  char i;

  for (i = 0; i < 2; i++)
  {
    if (key[i])
    {
      if (millis() - key_time[i] > 100)
      {
        key_time[i] = millis();
        key_action[i] = true;
      }
    }
  }

  for (i = 2; i < 4; i++)
  {
    if (key[i])
    {
      if (millis() - key_time[i] > 400)
      {
        key_time[i] = millis();
        key_action[i] = true;
      }
    }
  }  
}

void action()
{
  if ((key_action[0]) && (rpm > 0.1) && (!info) && (!control_enable))
  {
    key_action[0] = false;
    if (rpm >= 30) rpm--;
    else rpm = rpm - 0.1;
    refresh_speed = true;
  }

  if ((key_action[1]) && (rpm < max_rpm) && (!info) && (!control_enable))
  {
    key_action[1] = false;
    if (rpm >= 30) rpm++;
    else rpm = rpm + 0.1;
    refresh_speed = true;
  }

  if (key_action[2] && (!info))
  {
    key_action[2] = false;
    change_dir();
    refresh_dir = true;
  }

  if (key_action[3] && (!control_enable))
  {
    key_action[3] = false;
    if (!info)
     {
      refresh_run = true;   
      if (!run)  start = true;
      else run = false;
     }      
    if (info)
    {
      info = false;
      refresh_disp = true;
    }
  }

  if (info && control_enable)
     {
      info = false;
      refresh_disp = true;
    } 

  if (control_enable) 
   {
    read_ext_control();
    if (ext_on && !ext_on_prev)
     {
      start=true;
      ext_on_prev=true;
     }
    if (!ext_on)
     {
      run=false;
      ext_on_prev=false;
     }     
   }
}

void accelerate()
{
  run = true;
  set_pins();
  int i;
  for (i = 1; i < rpm; i++)
  {
    speed = tic / (i);
    delay(3);
  }
  start = false;
}

void decelerate()
{
  int i;
  for (i = rpm; i > 10; i--)
  {
    speed = tic / i;
    delay(1);
  }
  run = false;
}

void disp()
{
  if (info)
  {
    if (refresh_info)
    {
      lcd.setCursor(0, 0);
      lcd.print(info_1);
      lcd.setCursor(0, 1);
      lcd.print(info_2);
      refresh_info = false;
    }
  }
  else
  {
/*    if (ext_on)
    {
      lcd.setCursor(0, 0);
      lcd.print("R");
     }
    else
     {
      lcd.setCursor(0, 0);
      lcd.print(" ");
     }
*/
    if (refresh_disp)
    {
      lcd.clear();
      lcd.setCursor(1, 1);
      lcd.write(byte(0));
      lcd.setCursor(5, 1);
      lcd.write(byte(1));
      lcd.setCursor(8, 1);
      lcd.print("<=>");
      lcd.setCursor(0, 0);
      lcd.print("Sp:");
      refresh_disp = false;
    }

    if (refresh_speed)
    {
      dtostrf(rpm, 3, 1, rpm_str);
//      dtostrf(external, 4, 0, adc_str);
      lcd.setCursor(3, 0);
      lcd.print(rpm_str);
      lcd.print("RPM ");
      if (rpm<10) lcd.print(" ");
//      lcd.print(" ");
//      lcd.setCursor(7, 0);
//      lcd.print(adc_str);
      refresh_speed = false;
    }

    if (refresh_dir)
    {
      lcd.setCursor(12, 0);
      if (dir) lcd.print("<=");
      else lcd.print("=>");
      refresh_dir = false;
    }

    if (refresh_cont)
    {
      if (control_enable)
      {
        lcd.setCursor(15, 0);
        lcd.print("*");
      }
      else
      {
        lcd.setCursor(15, 0);
        lcd.print(" ");
      }
      refresh_cont = false;
    }

   if (refresh_run)
    {
      if (run)
      {
        lcd.setCursor(13, 1);
        lcd.print("STP");
      }
      else
      {
        lcd.setCursor(13, 1);
        lcd.print("RUN");
      }
    }   
  }
}

void change_dir()
{
  if (run)
  {
    decelerate();
    dir = !dir;
    start = true;
  }
  else dir = !dir;
}

void set_pins()
{
  if (run) digitalWrite(EN_PIN , LOW);
  else  digitalWrite(EN_PIN , HIGH);
  if (dir) digitalWrite(DIR_PIN , LOW);
  else  digitalWrite(DIR_PIN , HIGH);
}

void calculate()
{
  if (control_enable && (external != external_old))
  {
    refresh_speed = true;
    refresh_run = true;
    rpm = (max_rpm * external) / 1024;
    if (rpm == 0) rpm = 0.1;
    external_old = external;
  }
  speed = tic / rpm;
}

void loop()
{
  disp();
  if (start) accelerate();
  key_read();
  key_check();
  action();
  calculate();
  set_pins();
}

ISR (TIMER2_COMPA_vect)
{
  if (count > 0)
  {
    count--;
    OCR2A = 0x80;
  }
  else
  {
    digitalWrite(STEP_PIN , HIGH);
    count = speed >> 7;
    OCR2A = speed % 0x80;
    digitalWrite(STEP_PIN , LOW);
  }
}


